require 'cgi'
require 'json'

# A parsed webproxy configuration
#
# This class unifies different tiller inputs into a consolidated configuration
# object containing the webproxy to use (if any).
class WebproxyConfiguration
    # Constructor
    #
    # @param scheme [String] HTTP or HTTPS
    # @param fqdn [String] FQDN or IP on which the webproxy is reachable
    # @param port [Integer] Port on which the webproxy is reachable
    # @param credentials [WebproxyAuth|nil] Optional credentials
    # @raise [Error] The provided values are not valid
    def initialize(scheme, fqdn, port, credentials)
        @scheme = scheme.to_s.downcase
        raise "Invalid webproxy scheme #{scheme}" unless /^https?$/i =~ @scheme
        @fqdn = fqdn.to_s
        raise "Invalid webproxy fqdn #{fqdn}" unless
            /^[a-zA-Z0-9._-]+$/ =~ @fqdn
        @port = port
        raise "Invalid webproxy port #{port}" unless
            @port.is_a?(Integer) && @port > 0 && @port < 65536
        @credentials = credentials
        raise "Invalid webproxy credentials" unless
            @credentials.nil? || @credentials.is_a?(WebproxyAuth)
    end

    # @return [String] The proxy configuration endpoint (fqdn+port)
    def endpoint
        "#{@fqdn}:#{@port}"
    end

    # @return [String] The proxy configuration URL (scheme+endpoint+auth)
    def url
        auth = @credentials.nil? ? '' : "#{@credentials.url_auth}@"
        "#{@scheme}://#{auth}#{self.endpoint}"
    end

    # @return [Boolean] true if the config scheme is HTTPS, false otherwise
    def encrypted
        @scheme == 'https'  # we down-case this in the constructor already
    end

    # @return [Boolean] true if the config contains credentials, false otherwise
    def authenticated
        ! @credentials.nil?
    end

    # @return [String] Username used for authentication
    # @raise [Error] The config does not use authentication. Check `authenticated`
    def username
        raise "Invalid call to username" unless self.authenticated
        @credentials.username
    end

    # @return [String] Password used for authentication
    # @raise [Error] The config does not use authentication. Check `authenticated`
    def password
        raise "Invalid call to password" unless self.authenticated
        @credentials.password
    end

    # Copy-constructor, replacing the endpoint with the given values
    #
    # @param scheme [String] HTTP or HTTPS
    # @param fqdn [String] FQDN or IP on which the webproxy is reachable
    # @param port [Integer] Port on which the webproxy is reachable
    # @return [String] Password used for authentication
    # @raise [Error] The config does not use authentication. Check `authenticated`
    def for_endpoint(scheme, fqdn, port)
        WebproxyConfiguration.new(scheme, fqdn, port, @credentials)
    end
end

# Class holding credentials for webproxy authentication
class WebproxyAuth
    # Constructor
    #
    # @param username [String] Username for authentication
    # @param password [String] Password for authentication
    def initialize(username, password)
        @username = username.to_s
        @password = password.to_s
    end

    # @return [String] The credentials as they can be embedded in a URL (before the @<domain>)
    def url_auth
        # URL-encode username and password in case they contain special characters that are
        # used as tokenizers (e.g., : or @)
        "#{CGI.escape(@username)}:#{CGI.escape(@password)}"
    end

    # @return [String] Username used for authentication
    def username
        @username
    end

    # @return [String] Password used for authentication
    def password
        @password
    end
end

# Factory class for parsing the webconfig configured in tiller (if any)
#
# Unlike for YAML (which is a Lastline addition to tiller), tiller allows parsing only
# one single JSON environment variable as input. NSX deployments use JSON for loading
# any user configuration into kubernetes, so we have to be flexible in loading user
# configurations.
#
# Furthermore, the NSX webproxy configuration is a JSON blob that does not fit our
# tiller configuration schema, so we want to have a central way to parse, re-structure,
# validate, and return inputs, so all templates can consume this validated config.
class WebproxyConfigurationFactory
    # Factory method returning the webconfig configured in tiller (if any)
    #
    # @param tiller_webproxy_config [Hash] Tiller webproxy configuration hash to parse
    # @return [WebproxyConfiguration|nil] Webproxy configuration (if set), nil otherwise
    # @raise [Error] Invalid configuration found
    def self.from_tiller_config(tiller_webproxy_config)
        return nil unless tiller_webproxy_config['enabled']
        return self.from_tiller_nsx_config_json_env(
            tiller_webproxy_config['nsx_config_json_env']
        ) if tiller_webproxy_config['enabled'] == 'nsx_config_json_env'
        raise "Invalid webproxy enabled #{tiller_webproxy_config['enabled']}" unless
            tiller_webproxy_config['enabled'] == true
        self.from_default_tiller_tree(tiller_webproxy_config)
    end

    # Factory method returning the webconfig specified as NSX JSON config blob
    #
    # @param env_variable [String] The name of the environment variable holding the NSX webproxy
    #     JSON config (if any)
    # @return [WebproxyConfiguration|nil] Webproxy configuration (if set), nil otherwise
    # @raise [Error] Invalid configuration found
    def self.from_tiller_nsx_config_json_env(env_variable)
        nsx_json_config = ENV[env_variable]
        return nil if nsx_json_config.nil? || nsx_json_config.strip.empty?

        config = JSON.parse(nsx_json_config)
        credentials = (
            config['userName'].nil? || config['userName'].empty? ||
            config['password'].nil? || config['password'].empty?
        ) ? nil : WebproxyAuth.new(config['userName'], config['password'])
        WebproxyConfiguration.new(
            # NP-4942: NSX common-agent fails to propagate scheme into the k8s JSON object, so we
            # may not get a scheme here and default to HTTP
            config.key?('scheme') ? config['scheme'] : 'http',
            config['hostName'],
            config['port'],
            credentials,
        )
    end

    # Factory method returning the webconfig specified as part of the tiller config tree
    #
    # @param tiller_webproxy_config [Hash] Tiller webproxy configuration hash
    # @return [WebproxyConfiguration] Webproxy configuration
    # @raise [Error] Invalid configuration found
    def self.from_default_tiller_tree(webproxy_config)
        auth_config = webproxy_config['credentials']
        credentials = (
            auth_config['username'].nil? || auth_config['username'].empty? ||
            auth_config['password'].nil? || auth_config['password'].empty?
        ) ? nil : WebproxyAuth.new(auth_config['username'], auth_config['password'])
        WebproxyConfiguration.new(
            webproxy_config['scheme'],
            webproxy_config['fqdn'],
            webproxy_config['port'],
            credentials,
        )
    end
end
